@@ -1,9 +1,12 @@
+import time
 from typing import Tuple
 import requests
 from requests.auth import HTTPBasicAuth
 import fastavro
 from enum import StrEnum
 import json
+
+from cc_clients_python_lib.http_status import HttpStatus
  
 
 __copyright__  = "Copyright (c) 2025 Jeffrey Jonathan Jennings"
@@ -54,8 +57,7 @@ class SchemaRegistryClient():
             dict:   The subject's latest schema.  Otherwise, an empty dict is returned.
         """
         # Send a GET request to get the subject's latest schema.
-        response = requests.get(url=f"{self.schema_registry_url}/subjects/{subject_name}/versions/latest", 
-                                auth=HTTPBasicAuth(self.api_key, self.api_secret))
+        response = requests.get(url=f"{self.schema_registry_url}/subjects/{subject_name}/versions/latest", auth=HTTPBasicAuth(self.api_key, self.api_secret))
  
         try:
             # Raise HTTPError, if occurred.
@@ -66,30 +68,69 @@ class SchemaRegistryClient():
         except requests.exceptions.RequestException as e:
             return response.status_code, f"Error retrieving subject '{subject_name}': {e}",  response.json() if response.content else {}
        
-    def register_topic_subject_schema(self, subject_name: str, schema_str: str) -> Tuple[int, str, int]:
+    def register_topic_subject_schema(self, subject_name: str, subject_schema, force: bool = False) -> Tuple[int, str, int]:
         """This function submits a RESTful API call to register the subject's schema.
  
         Arg(s):
             schema_registry_confg (dict):  The Schema Registry Cluster configuration.
             subject_name (str):            The Kafka topic subject name.
-            schema_str (str):              The subject's new schema to be registered.
+            subject_schema (any):          The subject's new schema to be registered.
+            force (bool):                  (Optional)  A flag to force the registration of the schema, by setting the Compatibility
+                                           to NONE, register schema, and then set back to orginal setting.
  
         Returns:
             int:  HTTP Status Code.
             str:  HTTP Error, if applicable.
             int:  The schema ID of the newly created schema.
         """
+        if force:
+            # Get the current Kafka topic subject compatibility level.
+            http_status_code, http_error_message, current_compatibility_level = self.get_topic_subject_compatibility_level(subject_name)
+            match http_status_code:
+                case HttpStatus.OK:
+                    pass
+                case HttpStatus.NOT_FOUND:
+                    http_status_code, http_error_message, current_compatibility_level = self.get_global_topic_subject_compatibility_level()
+                    if current_compatibility_level == CompatibilityLevel.UNASSIGNED:
+                        return http_status_code, f"Error retrieving the global compatibility level because {http_error_message}.", -1
+                case _:
+                    return http_status_code, f"Error retrieving the current compatibility level because {http_error_message}.", -1
+        
+            # Set the topic subject compatibility setting to "NONE", so the existing subject's latest schema
+            # can be overwritten.
+            http_status_code, http_error_message = self.set_topic_subject_compatibility_level(subject_name, CompatibilityLevel.NONE)
+            if http_status_code != HttpStatus.OK:
+                return http_status_code, f"Error setting the current compatibility level because {http_error_message}.", -1
+    
+        # Convert the Avro schema to a string.
+        if isinstance(subject_schema, dict):
+            schema_str, error_message = self.convert_avro_schema_into_string(subject_name, subject_schema)
+            if error_message != "":
+                return HttpStatus.INTERNAL_SERVER_ERROR, error_message, -1
+        else:
+            schema_str = subject_schema        
+   
         # Send a POST request to register the schema.
         response = requests.post(url=f"{self.schema_registry_url}/subjects/{subject_name}/versions",
-                                 json={"scheam": schema_str},
+                                 json={"schema": schema_str},
                                  auth=HTTPBasicAuth(self.api_key, self.api_secret))
  
         try:    
             # Raise HTTPError, if occurred.
             response.raise_for_status()
- 
-            # Return the new schema ID of the newly registered schema.
-            return response.status_code, response.text, response.json().get("id")
+
+            schema_id = response.json().get("id")
+            schema_registration_result = response.text
+
+            if force:
+                # Restore the topic subject compatibility level.
+                http_status_code, http_error_message = self.set_topic_subject_compatibility_level(subject_name, current_compatibility_level)
+                if http_status_code == HttpStatus.OK:
+                    return response.status_code, schema_registration_result, schema_id
+                else:
+                    return http_status_code, http_error_message, -1
+            else:
+                return response.status_code, schema_registration_result, schema_id
         except requests.exceptions.RequestException as e:
             return response.status_code, f"Error registering subject '{subject_name}': {e}", -1
    
@@ -97,8 +138,8 @@ class SchemaRegistryClient():
         """This function submits a RESTful API call to set the topic subject compatibility level.
  
         Arg(s):
-            schema_registry_confg (dict):                   The Schema Registry Cluster configuration.
-            subject_name (str):                             The Kafka topic subject name.
+            schema_registry_confg (dict):               The Schema Registry Cluster configuration.
+            subject_name (str):                         The Kafka topic subject name.
             compatibility_level (CompatibilityLevel):   The compatibility setting.
  
         Returns:
@@ -136,10 +177,7 @@ class SchemaRegistryClient():
  
         try:
             # Send a GET the Topic Subject compatibility setting level.
-            response = requests.get(
-                endpoint,
-                auth=HTTPBasicAuth(self.api_key, self.api_secret)
-            )
+            response = requests.get(endpoint, auth=HTTPBasicAuth(self.api_key, self.api_secret))
  
             # Raise HTTPError, if occurred.
             response.raise_for_status()
@@ -161,8 +199,7 @@ class SchemaRegistryClient():
             compatibility_level:  The Topic Subject compatibility level.
         """
         # Send a GET the Topic Subject compatibility setting level.
-        response = requests.get(url=f"{self.schema_registry_url}/config",
-                                auth=HTTPBasicAuth(self.api_key, self.api_secret))
+        response = requests.get(url=f"{self.schema_registry_url}/config", auth=HTTPBasicAuth(self.api_key, self.api_secret))
 
         try:
             # Raise HTTPError, if occurred.
@@ -201,35 +238,52 @@ class SchemaRegistryClient():
             kafka_topic_name (str):  The Kafka topic name of the key schema subject.
  
         Returns:
-            int:                  HTTP Status Code.
-            str:                  HTTP Error, if applicable.
+            int: HTTP Status Code.
+            str: HTTP Error, if applicable.
         """
         # Send a DELETE to perform a soft-delete of all version of the schema.
-        response = requests.delete(url=f"{self.schema_registry_url}/subjects/{kafka_topic_name}-key",
-                                   auth=HTTPBasicAuth(self.api_key, self.api_secret))
+        delete_response = requests.delete(url=f"{self.schema_registry_url}/subjects/{kafka_topic_name}-key", auth=HTTPBasicAuth(self.api_key, self.api_secret))
  
         try:
             # Raise HTTPError, if occurred.
-            response.raise_for_status()
+            delete_response.raise_for_status()
         except requests.exceptions.RequestException as e:
-            return response.status_code, f"Delete topic schema subject soft-delete failed because {e}"
-        
-        # The Confluent Schema Registry endpoint to delete the topic schema subject hard-delete of all versions registered.
-        endpoint = f"{self.schema_registry_url}/subjects/{kafka_topic_name}-key?permanent=true"
+            return delete_response.status_code, f"Delete topic schema subject soft-delete failed because {e} and the response returned was {delete_response.text}"
  
         try:
             # Send a DELETE to perform a hard-delete of all version of the schema.
-            response = requests.delete(
-                endpoint,
-                auth=HTTPBasicAuth(self.api_key, self.api_secret)
-            )
+            delete_response = requests.delete(f"{self.schema_registry_url}/subjects/{kafka_topic_name}-key?permanent=true", auth=HTTPBasicAuth(self.api_key, self.api_secret))
  
             # Raise HTTPError, if occurred.
-            response.raise_for_status()
+            delete_response.raise_for_status()
 
-            return response.status_code, response.text
+            retry = 0
+            max_retries = 3
+            retry_delay_in_seconds = 5
+
+            while retry < max_retries:
+                # Send a GET request to get the subject's latest schema.
+                get_response = requests.get(url=f"{self.schema_registry_url}/subjects/{kafka_topic_name}-key/versions/latest", auth=HTTPBasicAuth(self.api_key, self.api_secret))
+        
+                try:
+                    # Raise HTTPError, if occurred.
+                    get_response.raise_for_status()
+
+                    retry += 1
+                    if retry == max_retries:
+                        return get_response.status_code, f"Max retries exceeded.  Fail to check if the subject '{kafka_topic_name}-key' exist because the response is {get_response.text}.  But not sure if the Subject Schema Key is deleted or not."
+                    else:
+                        time.sleep(retry_delay_in_seconds)
+                except requests.exceptions.RequestException as e:
+                    retry += 1
+                    if retry == max_retries:
+                        return get_response.status_code, f"Max retries exceeded.  Fail to check if the subject '{kafka_topic_name}-key' exist because {e}, and the response is {get_response.text}.  But not sure if the Subject Schema Key is deleted or not."
+                    elif get_response.status_code == HttpStatus.NOT_FOUND:
+                        return HttpStatus.OK, delete_response.text
+                    else:
+                        time.sleep(retry_delay_in_seconds)
         except requests.exceptions.RequestException as e:
-            return response.status_code, f"Delete topic key schema subject hard-delete failed because {e}"
+            return delete_response.status_code, f"Delete topic key schema subject hard-delete failed because {e} and the response returned was {delete_response.text}"
         
     def delete_kafka_topic_value_schema_subject(self, kafka_topic_name) -> Tuple[int, str]:
         """This function submits a RESTful API call to delete the topic value schema subject.
@@ -238,30 +292,52 @@ class SchemaRegistryClient():
             kafka_topic_name (str):  The Kafka topic name of the value schema subject.
  
         Returns:
-            int:                  HTTP Status Code.
-            str:                  HTTP Error, if applicable.
+            int: HTTP Status Code.
+            str: HTTP Error, if applicable.
         """
         # Send a DELETE to perform a soft-delete of all version of the schema.
-        response = requests.delete(url=f"{self.schema_registry_url}/subjects/{kafka_topic_name}-value",
-                                   auth=HTTPBasicAuth(self.api_key, self.api_secret))
+        delete_response = requests.delete(url=f"{self.schema_registry_url}/subjects/{kafka_topic_name}-value", auth=HTTPBasicAuth(self.api_key, self.api_secret))
  
         try:
             # Raise HTTPError, if occurred.
-            response.raise_for_status()
+            delete_response.raise_for_status()
         except requests.exceptions.RequestException as e:
-            return response.status_code, f"Delete topic schema subject soft-delete failed because {e}"
+            return delete_response.status_code, f"Delete topic schema subject soft-delete failed because {e} and the response returned was {delete_response.text}"
         
         # The Confluent Schema Registry endpoint to delete the topic schema subject hard-delete of all versions registered.
         # Send a DELETE to perform a hard-delete of all version of the schema.
-        response = requests.delete(url=f"{self.schema_registry_url}/subjects/{kafka_topic_name}-value?permanent=true",
-                                   auth=HTTPBasicAuth(self.api_key, self.api_secret))
+        delete_response = requests.delete(url=f"{self.schema_registry_url}/subjects/{kafka_topic_name}-value?permanent=true", auth=HTTPBasicAuth(self.api_key, self.api_secret))
  
         try:
             # Raise HTTPError, if occurred.
-            response.raise_for_status()
+            delete_response.raise_for_status()
 
-            return response.status_code, response.text
+            retry = 0
+            max_retries = 3
+            retry_delay_in_seconds = 5
+
+            while retry < max_retries:
+                # Send a GET request to get the subject's latest schema.
+                get_response = requests.get(url=f"{self.schema_registry_url}/subjects/{kafka_topic_name}-value/versions/latest", auth=HTTPBasicAuth(self.api_key, self.api_secret))
+        
+                try:
+                    # Raise HTTPError, if occurred.
+                    get_response.raise_for_status()
+
+                    retry += 1
+                    if retry == max_retries:
+                        return get_response.status_code, f"Max retries exceeded.  Fail to check if the subject '{kafka_topic_name}-value' exist because the response is {get_response.text}.  But not sure if the Subject Schema Value is deleted or not."
+                    else:
+                        time.sleep(retry_delay_in_seconds)
+                except requests.exceptions.RequestException as e:
+                    retry += 1
+                    if retry == max_retries:
+                        return get_response.status_code, f"Max retries exceeded.  Fail to check if the subject '{kafka_topic_name}-value' exist because {e}, and the response is {get_response.text}.  But not sure if the Subject Schema Value is deleted or not."
+                    elif get_response.status_code == HttpStatus.NOT_FOUND:
+                        return HttpStatus.OK, delete_response.text
+                    else:
+                        time.sleep(retry_delay_in_seconds)
         except requests.exceptions.RequestException as e:
-            return response.status_code, f"Delete topic value schema subject hard-delete failed because {e}"
+            return delete_response.status_code, f"Delete topic value schema subject hard-delete failed because {e} and the response returned was {delete_response.text}"
 
  
\ No newline at end of file
